#ifndef lint
static char rcsid[] = "@(#)$Header: kbinfo.c,v 1.15 92/08/21 18:35:07 leres Exp $ (LBL)";
#endif

/*
 * kbinfo - dump statistics or config from a kbox.
 *	kbinfo [-c] hostname
 *
 * This program sends a debug message to a Kinetics Fastpath
 * running the KIP code, asking for various device and error
 * stats kept by the box.  If any reply is received, it is
 * printed to stdout.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <netinet/in.h>

#include <signal.h>
#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <netdb.h>
#define iaddr_t u_long
#include "cmdmacro.h"
#include "gwctl.h"

#define KBINFO_RETRIES 2	/* number of retries */
#define KBINFO_TIMEOUT 2	/* seconds between retries */

extern char *malloc(), *calloc();
extern char *sprintf();
char *ip_print();
char *at_print();
char *intoa();
char *etoa();

int s;
struct sockaddr_in fsin = {AF_INET};	/* foreign sockaddr_in */
int fsinlen;

int debug;
int doconfig;
int nflag;

static char srec_version[512];

main(argc, argv)
	int argc;
	char *argv[];
{
	register struct hostent *host;
	register int n;
	u_long iaddr;
	register struct gwdb *dp;
	register struct gwdb_Stats *sp;
	char buf[512];
	char *usage = "usage: kbinfo [-c] [-n] hostname\n";

	for (argc--, argv++; argc > 0; argc--, argv++) {
		if (argv[0][0] == '-') {
			switch (argv[0][1]) {
			case 'd':
				++debug;
				break;
			case 'c':
				++doconfig;
				break;
			case 'n':
				++nflag;
				break;
			default:
				fprintf(stderr, usage);
				exit (1);
			}
			continue;
		} else {
			if (isdigit(**argv)) {
				if ((iaddr = inet_addr(*argv)) == -1
				    || iaddr == 0) {
					fprintf(stderr, "bad ipaddress %s\n",
					    *argv);
					exit (1);
				}
			}
			else {
				if ((host = gethostbyname(*argv)) == 0) {
					fprintf(stderr, "bad hostname %s\n",
					    *argv);
					exit (1);
				}
				bcopy(host->h_addr, (caddr_t) &iaddr,
				    sizeof iaddr);
			}
			fsin.sin_addr.s_addr = iaddr;
		}
	}
	if (fsin.sin_addr.s_addr == 0) {
		fprintf(stderr, usage);
		exit (1);
	}
	if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("kbinfo: socket");
		exit(1);
	}
	fsin.sin_port = gwdbPort;
#ifndef sparc
	dp = (struct gwdb *)malloc(sizeof(*dp));
#else
	/* Make enough room so that we don't have do do fancy length calcs */
	dp = (struct gwdb *)malloc(2 * sizeof(*dp));
#endif
	if (dp == 0) {
		fprintf(stderr, "kbinfo: gwdb malloc() failed\n");
		exit(1);
	}
	bzero((caddr_t)dp, sizeof(*dp));
	dp->magic = gwdbMagic;
	dp->op = gwdbStats;
	dp->seq = 1;

	/*
	 * Retry so we can detect dropped packets and FP-2's (which don't
	 * answer this kind of request).
	 */

	/* Send the packet and wait for a reply (with timeout) */
	n = sendrecv((caddr_t) dp, sizeof(*dp),
	    (caddr_t) dp, sizeof(*dp), "stats");
	if (n < 0)
		exit(1);

	/* Complain if it's mis-sized */
	if (n < sizeof(*dp) || dp->count < sizeof(*sp)) {
		fprintf(stderr,
		    "kbinfo: bogus reply of %d (%d) bytes from %s\n",
		    n, dp->count, intoa(fsin.sin_addr.s_addr));
		exit(2);
	}

	/* print the ident info */
	sp = (struct gwdb_Stats *)dp->data;
	bcopy((caddr_t)&sp->version.fpv_model, buf, 4);
	buf[4] = 0;
	printf("%s: %s model #%d/%g running v%g proms.\n",
		ip_print(fsin.sin_addr.s_addr), buf,
		sp->ident.fpid_model, ((float)sp->ident.fpid_version) / 100.,
		((float)sp->version.fpv_version) / 100.);

	printf("  name %s, sn#%d, state %d, up ",
		sp->ident.fpid_name, sp->ident.fpid_sn,
		sp->ident.fpid_state);
	print_uptime(sp->uptime);
	putchar('\n');


	if (doconfig) {
		/* if there's a version string in the stats info, snarf it */
		if (*(char *)(sp + 1))
			bcopy((caddr_t)(sp + 1), srec_version,
			    sizeof(dp->data) - sizeof(*sp));

		/* get the config info */
		bzero((caddr_t)dp, sizeof(*dp));
		dp->magic = gwdbMagic;
		dp->op = gwdbState;
		dp->seq = 1;

		/* Send the packet and wait for a reply (with timeout) */
		n = sendrecv((caddr_t) dp, sizeof(*dp),
		    (caddr_t) dp, sizeof(*dp), "state");
		if (n < 0)
			exit(1);

		/* Complain if it's mis-sized */
		if (n < sizeof(*dp) || dp->count < sizeof(*sp)) {
			fprintf(stderr,
			    "kbinfo: bogus reply of %d (%d) bytes from %s\n",
			    n, dp->count, intoa(fsin.sin_addr.s_addr));
			exit(2);
		}
		print_conf((struct gwdb_State *)dp->data);
	}
	else
		print_stats(sp);

	exit(0);
	/*NOTREACHED*/
}

#ifndef FD_SETSIZE
	/* Gag. */
#define FD_SET(n, p)	((p)->fds_bits[0] |= (1<<(n)))
#define FD_CLR(n, p)	((p)->fds_bits[0] &= ~(1<<(n)))
#define FD_ISSET(n, p)	((p)->fds_bits[0] & (1<<(n)))
#define FD_ZERO(p)	((p)->fds_bits[0] = 0)
#endif

sendrecv(sbuf, slen, rbuf, rlen, str)
	caddr_t sbuf;
	int slen;
	caddr_t rbuf;
	int rlen;
	char *str;
{
	struct timeval timeout;
	register int n, i;
	fd_set readfds;

	timeout.tv_sec = KBINFO_TIMEOUT;
	timeout.tv_usec = 0;
	for (i = KBINFO_RETRIES; i > 0; i--) {
		if (sendto(s, sbuf, slen, 0, &fsin, sizeof(fsin)) < 0) {
			fprintf(stderr, "kbinfo: %s", str);
			perror("sendto");
			return(-1);
		}
		FD_ZERO(&readfds);
		FD_SET(s, &readfds);
		n = select(s+1, &readfds, (fd_set*)0, (fd_set*)0, &timeout);
		if (n < 0) {
			fprintf(stderr, "kbinfo: %s", str);
			perror("select");
			return(-1);
		}
		if (FD_ISSET(s, &readfds))
			break;

	}
	if (i <= 0) {
		(void)fprintf(stderr, "kbinfo: %s timeout\n", str);
		return(-1);
	}

	fsinlen = sizeof(fsin);
	n = recvfrom(s, rbuf, rlen, 0, (caddr_t)&fsin, &fsinlen);
	if (n < 0) {
		fprintf(stderr, "kbinfo: %s", str);
		perror("recv");
		return(-1);
	}
	return(n);
}

print_stats(sp)
	struct gwdb_Stats *sp;
{
	printf("  Applebus stats:\n");
	pnz(12, (u_int)sp->abstats.IntCount, "receive interrupts");
	pnzp(20, (u_int)sp->abstats.RcvCount, "good data packets received",
	    sp->abstats.IntCount);
	pnzp(20, (u_int)sp->abstats.BadCount, "bad packets",
	    sp->abstats.RcvCount);
	pnzp(20, (u_int)sp->abstats.CRCCount, "bad CRC", sp->abstats.RcvCount);
	pnzp(20, (u_int)sp->abstats.OvrCount, "receive overruns",
	    sp->abstats.RcvCount);
	pnzp(20, (u_int)sp->abstats.NoHandCnt, "no protocol handler for packet",
	    sp->abstats.RcvCount);
	pnzp(20, (u_int)sp->abstats.LenErrCnt,
	    "packet length didn't match header", sp->abstats.RcvCount);
	pnzp(20, (u_int)sp->abstats.NoDtaCount, "no packet after RTS",
	    sp->abstats.RcvCount);
	pnz(12, (u_int)sp->abstats.XmitCount, "data packets transmitted");
	pnzp(20, (u_int)sp->abstats.UndCount, "transmit underruns",
	    sp->abstats.XmitCount);
	pnzp(20, (u_int)sp->abstats.AbortCount, "abort interrupts",
	    sp->abstats.XmitCount);
	pnzp(20, (u_int)sp->abstats.CollsnCount, "collisions",
	    sp->abstats.XmitCount);
	pnzp(20, (u_int)sp->abstats.DeferCount, "defers to other packets",
	    sp->abstats.XmitCount);
	pnz(12, (u_int)sp->abstats.RandomCTS, "unsolicted CTS");
	pnz(12, (u_int)sp->abstats.SpurCount, "spurious interrupts");

	printf("  Ethernet stats:\n");
	pnz(12, (u_int)sp->iestats.fpi_stats[IN_PACKETS], "packets in");
	pnzp(20, (u_int)sp->iestats.fpi_stats[SCB_CRCERRS], "crc errors",
	    sp->iestats.fpi_stats[IN_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[SCB_ALNERRS], "aln errors",
	    sp->iestats.fpi_stats[IN_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[SCB_RSCERRS], "rsc errors",
	    sp->iestats.fpi_stats[IN_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[SCB_OVRNERRS], "ovrn errors",
	    sp->iestats.fpi_stats[IN_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[OVERSIZE_FRAME],
	    "oversize frames", sp->iestats.fpi_stats[IN_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[SPUR_RU_READY],
	    "spurious ru rdys", sp->iestats.fpi_stats[IN_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[SPUR_CU_ACTIVE],
	    "spurious cu actives", sp->iestats.fpi_stats[IN_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[RCV_RESTARTS], "receive restarts",
	    sp->iestats.fpi_stats[IN_PACKETS]);
	pnz(12, (u_int)sp->iestats.fpi_stats[OUT_PACKETS], "packets out");
	pnzp(20, (u_int)sp->iestats.fpi_stats[BAD_TRANSMIT], "bad transmits",
	    sp->iestats.fpi_stats[OUT_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[XMIT_COLLSN], "xmit collsn",
	    sp->iestats.fpi_stats[OUT_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[XMIT_DEFER], "xmit defer",
	    sp->iestats.fpi_stats[OUT_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[XMIT_DMA_UNDERRUNS],
	    "xmit dma underrun", sp->iestats.fpi_stats[OUT_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[XMIT_MAX_COLLSN],
	    "xmit max collsn", sp->iestats.fpi_stats[OUT_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[XMIT_NO_CARRIER],
	    "xmit no carrier", sp->iestats.fpi_stats[OUT_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[XMIT_NO_CPS], "xmit no cps",
	    sp->iestats.fpi_stats[OUT_PACKETS]);
	pnzp(20, (u_int)sp->iestats.fpi_stats[XMIT_NO_SQE], "xmit no sqe",
	    sp->iestats.fpi_stats[OUT_PACKETS]);
	pnz(12, (u_int)sp->iestats.fpi_stats[SPUR_UNKNOWN], "spurious unknown");
	pnz(12, (u_int)sp->iestats.fpi_stats[BCAST_DROPS], "broadcast drops");
	pnz(12, (u_int)sp->iestats.fpi_stats[REINTERRUPTS], "reinterrupts");
	pnz(12, (u_int)sp->iestats.fpi_stats[ENET_BUF_REROUTE], "buf reroute");
	pnz(12, (u_int)sp->iestats.fpi_stats[ENET_PPN_DROP], "ppn drop");

#ifdef notdef
	printf("    DDP stats:\n");
	pnz(12, sp->ddpstats.fpddp_stats[BADATPSKT], "bad ATP socket");
	pnz(12, sp->ddpstats.fpddp_stats[BADECHO_REQ], "bad echo req");
	pnz(12, sp->ddpstats.fpddp_stats[BADHOPCNT], "bad hop cnt");
	pnz(12, sp->ddpstats.fpddp_stats[BADKLAPREQ], "bad KLAP req");
	pnz(12, sp->ddpstats.fpddp_stats[BADKLAPSKT], "bad KLAP socket");
	pnz(12, sp->ddpstats.fpddp_stats[BADRTMPREQ_SKT], "bad RTMP req socket");
	pnz(12, sp->ddpstats.fpddp_stats[BADRTMP_SKT], "bad RTMP socket");
	pnz(12, sp->ddpstats.fpddp_stats[BADZIPSKT], "bad ZIP socket");
	pnz(12, sp->ddpstats.fpddp_stats[BAD_DDPTYPE], "bad DDP type");
	pnz(12, sp->ddpstats.fpddp_stats[BAD_DESTNET], "bad dest. net");
	pnz(12, sp->ddpstats.fpddp_stats[COPY_BCAST], "broadcast copies");
	pnz(12, sp->ddpstats.fpddp_stats[NOBUF_CPBCAST], "no copy broadcasts");
	pnz(12, sp->ddpstats.fpddp_stats[ECHOLONG_DROP], "long echo drops");
	pnz(12, sp->ddpstats.fpddp_stats[ECHOSHORT_DROP], "short echo drops");
	pnz(12, sp->ddpstats.fpddp_stats[LAPBCAST_DROP], "LAP broadcast drops");
	pnz(12, sp->ddpstats.fpddp_stats[NBPLOOKUP_DROP], "NBP lookup drops");
	pnz(12, sp->ddpstats.fpddp_stats[RTMPREQ_DROP], "RTMT req drops");
	pnz(12, sp->ddpstats.fpddp_stats[SHORTDDP_DROP], "short DDP drops");
	pnz(12, sp->ddpstats.fpddp_stats[TOOLONG_DDP], "DDP too long");
	pnz(12, sp->ddpstats.fpddp_stats[TOOSHORT_DDP], "DDP too short");
#endif notdef

	printf("  Buffer stats (%d total @ %d bytes each):\n",
		sp->bufinfo.fpb_pbnavail, sp->bufinfo.fpb_bsize);
	pnz(12, (u_int)sp->bufinfo.fpb_pbndrops, "req. refused");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_FREE], "on free list");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_ABUS], "rcvd appletalk packets");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_ENET], "rcvd ethernet packets");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_ARP], "ARP packets");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_DATA], "general data");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_ERBF], "awaiting ethernet rcv");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_ETBF], "awaiting ethernet xmit");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_MALLOC], "malloced");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_UNUSED1], "UNUSED1");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_UNUSED2], "UNUSED2");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_UNUSED3], "UNUSED3");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_UNUSED4], "UNUSED4");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_UNUSED5], "UNUSED5");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_UNUSED6], "UNUSED6");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_UNUSED7], "UNUSED7");
	pnz(12, sp->bufinfo.fpb_pbntypes[PT_UNUSED8], "UNUSED8");
}


print_conf(sp)
	struct gwdb_State *sp;
{
	register int i;
	register struct fp_state *fp = &sp->state;

#ifdef sparc
	/* XXX stupid fp_state struct definition is misaligned... */
	{
	register char *cp;
	/* fps_options is preceded by 2 bytes of pad on the sparc */
	cp = (char *)&fp->fps_options;
	bcopy(cp - 2, cp, sizeof(*fp));

	/* So is fps_ipdefrouter. */
	cp = (char *)&fp->fps_ipdefrouter;
	bcopy(cp - 2, cp, sizeof(*fp));

	/* XXX cfgsum is misaligned too but this code doesn't look at it */
	}
#endif
	printf("  config'd for ");
	switch(fp->fps_type) {
	case 0:	printf("Appletalk Only"); break;
	case 1:	printf("Kinetics 'IP Subnet'"); break;
	case 2:	printf("KIP/KStar"); break;
	default: printf("unknown type %d", fp->fps_type); break;
	}
	if (fp->fps_autoboot)
		printf(", autoboot");
	if (fp->fps_autoconfig)
		printf(", autoconfig");
	if (fp->fps_pforce)
		printf(", pforce");
	if (fp->fps_options)
		printf(", options 0x%x", fp->fps_options);
	putchar('\n');

	printf("  config from \"%s\", srec \"%s\"",
		fp->fps_config, fp->fps_file);
	if (fp->fps_date) {
		struct tm *tp;
		struct timeb tbuf;
		char *cp;

		ftime(&tbuf);
		tp = localtime(&tbuf.time);
		fp->fps_date -= 0x7c25b080 - tbuf.timezone*60;
		if (tp->tm_isdst)
			fp->fps_date -= 60*60;
		cp = ctime(&fp->fps_date); cp[strlen(cp)-1] = 0;
		printf(" on %s", cp);
	}
	putchar('\n');

	if (srec_version[0])
		printf("  srec version \"%s\"\n", srec_version);

	printf("  ethernet %s", etoa(fp->fps_ether));
	if (fp->fps_etnet) {
		printf(", ET %s", at_print(fp->fps_etnet, fp->fps_ETnode));
		if (! fp->fps_ETvalid)
			putchar('*');
		if (fp->fps_etzone[0])
			printf(", zone \"%s\"", fp->fps_etzone);
	}
	putchar('\n');

	printf("  localtalk addresses: IP %s, AT %s",
	    ip_print((u_long)fp->fps_lp2), at_print(fp->fps_atnet,
	    fp->fps_node));
	if (! fp->fps_valid)
		putchar('*');
	if (fp->fps_atzone[0])
		printf(", zone \"%s\"", fp->fps_atzone);
	if (fp->fps_bridge)
		printf(", bridge @ %d", fp->fps_bridge);
	putchar('\n');

	printf("  ether addresses: IP %s, AT %s",
		ip_print(fp->fps_ipaddress),
		at_print(fp->fps_udpnet, fp->fps_udpnode));
	if (fp->fps_udpzone[0])
		printf(", zone \"%s\"", fp->fps_udpzone);
	putchar('\n');

	printf("  ether IP broadcast %s", intoa(fp->fps_ipbroadcast));
	printf(", netmask %s", intoa(fp->fps_ipsubmask));
	putchar('\n');

	printf("  admin %s", ip_print(fp->fps_ipkipserver));
	i = 1;
#define NEWLINEHACK if (i++ >= 2) { printf(",\n  "); i = 0; } else printf(", ")
	if (fp->fps_ipdefrouter) {
		++i;
		printf(", router %s", ip_print(fp->fps_ipdefrouter));
	}
	if (fp->fps_nameserver) {
		NEWLINEHACK;
		printf("nameserver %s", ip_print(fp->fps_nameserver));
	}
	if (fp->fps_fileserver) {
		NEWLINEHACK;
		printf("fileserver %s", ip_print(fp->fps_fileserver));
	}
	if (fp->fps_lp3) {
		NEWLINEHACK;
		printf("debug %s", ip_print((u_long)fp->fps_lp3));
	}
	if (i > 0)
		putchar('\n');

	if (fp->fps_ndynamics || fp->fps_nstatics) {
		printf("  IP address assignment:");
		if (fp->fps_ndynamics)
			printf("  %d dynamics", fp->fps_ndynamics);
		if (fp->fps_nstatics)
			printf("  %d statics", fp->fps_nstatics);
		putchar('\n');
	}
}


pnz(indent, val, msg)
	int indent;
	u_int val;
	char *msg;
{
	if (val)
		printf("%*u %s\n", indent, val, msg);
}


pnzp(indent, val, msg, val2)
	int indent;
	u_int val;
	char *msg;
	int val2;
{
	if (val) {
		if (val2)
			printf("%*u %s (%.1f%%)\n", indent, val, msg,
				(float)val / (float)val2 * 100.);
		else
			printf("%*u %s (??%%)\n", indent, val, msg);
	}
}


/*
 * Routine to return printable form of an ip number.
 */
char *
ip_print(addr)
	u_long addr;
{
	register struct hostent *hp;

	if (!nflag && addr &&
	    (hp = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET)))
		return(hp->h_name);

	return(intoa(addr));
}


/*
 * Routine to return printable form of an appletalk net/node number.
 */
#define HASHNAMESIZE 4096

struct hnamemem {
	int addr;
	char *name;
	struct hnamemem *nxt;
};

static struct hnamemem hnametable[HASHNAMESIZE];

char *
at_print(atnet, athost)
	u_short atnet;
	u_char athost;
{
	register struct hnamemem *tp, *tp2;
	register int i = (atnet << 8) | athost;
	char nambuf[256];
	static int first = 1;
	FILE *fp;

	/*
	 * if this is the first call, see if there's an appletalk
	 * number to name map file.
	 */
	if (!nflag && first && (fp = fopen("/etc/atalk.names", "r"))) {
		char line[256];
		int i1, i2, i3;

		while (fgets(line, sizeof(line), fp)) {
			if (line[0] == '\n' || line[0] == 0 || line[0] == '#')
				continue;
			if (sscanf(line, "%d.%d.%d %s", &i1, &i2, &i3,
			    nambuf) == 4)
				/* got a hostname. */
				i3 |= ((i1 << 8) | i2) << 8;
			else if (sscanf(line, "%d.%d %s", &i1, &i2,
					nambuf) == 3)
				/* got a net name */
				i3 = (((i1 << 8) | i2) << 8) | 255;
			else
				continue;

			for (tp = &hnametable[i3 & (HASHNAMESIZE-1)];
			    tp->nxt; tp = tp->nxt)
				;
			tp->addr = i3;
			tp->nxt = (struct hnamemem *)calloc (1, sizeof (*tp));
			i3 = strlen(nambuf) + 1;
			tp->name = strcpy(malloc((unsigned) i3), nambuf);
		}
		(void)fclose(fp);
		first = 0;
	}

	for (tp = &hnametable[i & (HASHNAMESIZE-1)]; tp->nxt; tp = tp->nxt)
		if (tp->addr == i)
			return (tp->name);

	/* didn't have the node name -- see if we've got the net name */
	i |= 255;
	for (tp2 = &hnametable[i & (HASHNAMESIZE-1)]; tp2->nxt; tp2 = tp2->nxt)
		if (tp2->addr == i) {
			tp->addr = (atnet << 8) | athost;
			tp->nxt = (struct hnamemem *)calloc (1, sizeof (*tp));
			(void)sprintf(nambuf, "%s.%d", tp2->name, athost);
			i = strlen(nambuf) + 1;
			tp->name = strcpy(malloc((unsigned) i), nambuf);
			return (tp->name);
		}

	tp->addr = (atnet << 8) | athost;
	tp->nxt = (struct hnamemem *)calloc (1, sizeof (*tp));
	(void)sprintf(nambuf, "%d.%d.%d", atnet >> 8, atnet & 0xff, athost);
	i = strlen(nambuf) + 1;
	tp->name = strcpy(malloc((unsigned) i), nambuf);

	return (tp->name);
}

#define MINUTE 60
#define HOUR (MINUTE*60)
#define DAY (HOUR*24)
#define WEEK (DAY*7)

print_uptime(up)
	register u_long up;
{
	if (up < MINUTE)
		printf("%u seconds", up);
	else if (up < 2*HOUR)
		printf("%u minutes", (up + MINUTE/2)/MINUTE);
	else if (up < 2*DAY)
		printf("%u hours", (up + HOUR/2)/HOUR);
	else if (up < 2*WEEK)
		printf("%u days", (up + DAY/2)/DAY);
	else
		printf("%u weeks", (up + WEEK/2)/WEEK);
}

/*
 * A faster replacement for inet_ntoa().
 */
char *
intoa(addr)
	u_long addr;
{
	register char *cp;
	register u_int byte;
	register int n;
	static char buf[sizeof(".xxx.xxx.xxx.xxx")];

#ifdef NTOHL
	NTOHL(addr);
#else
	addr = ntohl(addr);
#endif
	cp = &buf[sizeof buf];
	*--cp = '\0';

	n = 4;
	do {
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if (byte > 0) {
			*--cp = byte % 10 + '0';
			byte /= 10;
			if (byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);

	return cp + 1;
}

/*
 * A faster replacement for ether_ntoa().
 */
static char hex[] = "0123456789abcdef";

char *
etoa(ep)
	u_char *ep;
{
	register char *cp;
	register int n;
	static char buf[sizeof(":xx:xx:xx:xx:xx:xx:")];
	register u_int i, j;

	cp = buf;
	if (j = *ep >> 4)
		*cp++ = hex[j];
	*cp++ = hex[*ep++ & 0xf];
	for (i = 5; (int)--i >= 0;) {
		*cp++ = ':';
		if (j = *ep >> 4)
			*cp++ = hex[j];
		*cp++ = hex[*ep++ & 0xf];
	}
	*cp = '\0';
	return (buf);
}
